from ryvencore.Base import Base


class GUIBase:
    """Base class for GUI items that represent specific backend components"""

    # every frontend GUI object that represents some specific component from the backend
    # is stored there under the the global id of the represented component.
    # used for completing data (serialization)
    FRONTEND_COMPONENT_ASSIGNMENTS = {}  # component global id : GUI object

    @staticmethod
    def get_complete_data_function(session):
        """
        generates a function that searches through generated data by the backend and calls
        complete_data() on frontend components that represent them to add frontend data
        """

        def analyze(obj):
            """Searches recursively through obj and calls complete_data(obj) on associated
            frontend components (instances of GUIBase)"""

            if isinstance(obj, dict):
                GID = obj.get('GID')
                if GID is not None:
                    # find representative
                    comp = GUIBase.FRONTEND_COMPONENT_ASSIGNMENTS.get(GID)
                    if comp:
                        obj = session.threading_bridge__backend.run(
                            comp.complete_data, (obj,)
                        )  # run complete_data() (in frontend thread)

                # look for child objects
                for key, value in obj.items():
                    obj[key] = analyze(value)

            elif isinstance(obj, list):
                for i in range(len(obj)):
                    item = obj[i]
                    item = analyze(item)
                    obj[i] = item

            return obj

        return analyze

    def __init__(self, representing_component: Base = None):
        """parameter `representing` indicates representation of a specific backend component"""
        if representing_component is not None:
            GUIBase.FRONTEND_COMPONENT_ASSIGNMENTS[representing_component.GLOBAL_ID] = self

    # OVERRIDE
    def complete_data(self, data: dict) -> dict:
        """completes the data dict of the represented backend component by adding all frontend data"""
        return data



"""
These classes are used to abstract away threaded communication between GUI and  and backend components.
The backend can run trigger method executions in the frontend through the backend interface, and the
frontend can do the same respectively through the frontend interface.
Notice that both classes abstract away the concurrency of the threads such that they wait for some
container object to be set (the payload) with the return of the triggered method. This enables
calling methods of objects from another thread as if they lived in the same thread.
"""

from qtpy.QtCore import QObject, Signal

from .utils import Container, wait_until


class SessionThreadInterface_Backend(QObject):
    """lives in the session's thread"""

    run_signal = Signal(object, object, object)
    callback_signal = Signal(object)

    def __init__(self):
        super().__init__()

        self.frontend = SessionThreadInterface_Frontend(self)
        self.run_signal.connect(self.frontend.run_in_frontend)

    # run in frontend
    def run(self, target_method, args: tuple = (), wait=True):
        """run some method in frontend thread"""
        resp_container = Container()
        self.run_signal.emit(target_method, tuple(args), resp_container)

        if wait:
            wait_until(resp_container.is_set)
            return resp_container.payload

        return None

    # run in backend
    def run_in_backend(self, target_method, args: tuple, resp_container: Container, callback=None):

        ret = target_method(*args)

        if resp_container:
            resp_container.set(ret)

        if callback is not None:
            self.callback_signal.connect(callback)
            self.callback_signal.emit(ret)
            self.callback_signal.disconnect(callback)


class SessionThreadInterface_Frontend(QObject):
    """lives in the GUI thread and triggers method executions in the session thread"""

    run_signal = Signal(object, object, object, object)

    def __init__(self, backend_component: SessionThreadInterface_Backend):
        super().__init__()

        self.backend = backend_component
        self.run_signal.connect(self.backend.run_in_backend)

    # run in backend
    def run(self, target_method, args: tuple = (), wait=True, callback=None):
        """
        Runs some method in backend thread
        `wait` causes the frontend thread to wait until the execution finished.
        `callback` is optional and can be a function that gets called after the execution was completed.
        """

        resp_container = Container()
        self.run_signal.emit(target_method, tuple(args), resp_container, callback)

        if wait:
            wait_until(resp_container.is_set)
            return resp_container.payload

        return None

    # run in frontend
    def run_in_frontend(self, target_method, args: tuple = (), resp_container: Container = None):

        ret = target_method(*args)

        if resp_container:
            resp_container.set(ret)

    # CONVENIENCE METHODS
    def trigger_node_action(self, method, data=None):
        self.run(method, args=(data,) if data is not None else ())



from qtpy.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QScrollArea
from qtpy.QtCore import Qt, Signal

from .utils import search, sort_nodes, inc, dec
from ..node_list_widget.NodeWidget import NodeWidget

from statistics import median


class NodeListWidget(QWidget):
    """notice, that 'nodes' refers to node CLASSES here"""

    # SIGNALS
    escaped = Signal()
    node_chosen = Signal(object)

    def __init__(self, session):
        super().__init__()

        self.session = session
        self.nodes = []

        self.current_nodes = []             # currently selectable nodes
        self.active_node_widget_index = -1  # index of focused node widget
        self.active_node_widget = None      # focused node widget
        self.node_widgets = {}              # Node-NodeWidget assignments
        self._node_widget_index_counter = 0

        self.setup_UI()


    def setup_UI(self):

        self.main_layout = QVBoxLayout(self)
        self.main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.main_layout)

        # adding all stuff to the layout
        self.search_line_edit = QLineEdit(self)
        self.search_line_edit.setPlaceholderText('search for node...')
        self.search_line_edit.textChanged.connect(self.update_view)
        self.layout().addWidget(self.search_line_edit)


        self.list_scroll_area = QScrollArea(self)
        self.list_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setWidgetResizable(True)
        self.list_scroll_area.setContentsMargins(0, 0, 0, 0)

        self.list_scroll_area_widget = QWidget()
        self.list_scroll_area_widget.setContentsMargins(0, 0, 0, 0)
        self.list_scroll_area.setWidget(self.list_scroll_area_widget)

        self.list_layout = QVBoxLayout()
        self.list_layout.setContentsMargins(0, 0, 0, 0)
        self.list_layout.setAlignment(Qt.AlignTop)
        self.list_scroll_area_widget.setLayout(self.list_layout)

        self.layout().addWidget(self.list_scroll_area)

        self.update_view('')

        self.setStyleSheet(self.session.design.node_selection_stylesheet)

        self.search_line_edit.setFocus()


    def mousePressEvent(self, event):
        # need to accept the event, so the scene doesn't process it further
        QWidget.mousePressEvent(self, event)
        event.accept()


    def keyPressEvent(self, event):
        """key controls"""

        num_items = len(self.current_nodes)

        if event.key() == Qt.Key_Escape:
            self.escaped.emit()

        elif event.key() == Qt.Key_Down:
            self.set_active_node_widget_index(
                inc(self.active_node_widget_index, length=num_items)
            )
        elif event.key() == Qt.Key_Up:
            self.set_active_node_widget_index(
                dec(self.active_node_widget_index, num_items)
            )

        elif event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            if len(self.current_nodes) > 0:
                self.place_node(self.active_node_widget_index)
        else:
            event.setAccepted(False)


    def wheelEvent(self, event):
        # need to accept the event, so the scene doesn't process it further
        QWidget.wheelEvent(self, event)
        event.accept()


    def refocus(self):
        """focuses the search line edit and selects the text"""
        self.search_line_edit.setFocus()
        self.search_line_edit.selectAll()


    def update_list(self, nodes):
        """update the list of available nodes"""
        self.nodes = sort_nodes(nodes)
        self.update_view('')


    def update_view(self, search_text=''):
        if len(self.nodes) == 0:
            return

        search_text = search_text.lower()

        # remove all node widgets

        for i in reversed(range(self.list_layout.count())):
            self.list_layout.itemAt(i).widget().setParent(None)

        self.current_nodes.clear()

        self._node_widget_index_counter = 0

        # search
        sorted_distances = search(
            items={
                n: [n.title.lower()] + n.tags
                for n in self.nodes
            },
            text=search_text
        )

        # create node widgets
        cutoff = median(sorted_distances.values())
        for n, dist in sorted_distances.items():
            if search_text != '' and dist > cutoff:
                continue

            self.current_nodes.append(n)

            if self.node_widgets.get(n) is None:
                self.node_widgets[n] = self.create_node_widget(n)

            self.list_layout.addWidget(self.node_widgets[n])

        # focus on first result
        if len(self.current_nodes) > 0:
            self.set_active_node_widget_index(0)


    def create_node_widget(self, node):
        node_widget = NodeWidget(self, node)
        node_widget.custom_focused_from_inside.connect(self.node_widget_focused_from_inside)
        node_widget.setObjectName('node_widget_' + str(self._node_widget_index_counter))
        self._node_widget_index_counter += 1
        node_widget.chosen.connect(self.node_widget_chosen)

        return node_widget

    def node_widget_focused_from_inside(self):
        index = self.list_layout.indexOf(self.sender())
        self.set_active_node_widget_index(index)

    def set_active_node_widget_index(self, index):
        self.active_node_widget_index = index
        node_widget = self.list_layout.itemAt(index).widget()

        if self.active_node_widget:
            self.active_node_widget.set_custom_focus(False)

        node_widget.set_custom_focus(True)
        self.active_node_widget = node_widget
        self.list_scroll_area.ensureWidgetVisible(self.active_node_widget)


    def node_widget_chosen(self):
        index = int(self.sender().objectName()[self.sender().objectName().rindex('_')+1:])
        self.place_node(index)


    def place_node(self, index):
        node_index = index
        node = self.current_nodes[node_index]
        self.node_chosen.emit(node)
        self.escaped.emit()



# unwrapped backend classes
from ryvencore import InfoMsgs, NodeInputBP, NodeOutputBP, ExecConnection, dtypes, \
    LogsManager, Logger, Flow, DataConnection
from ryvencore.script_variables import VarsManager

# wrapped classes
from .Node import Node
from .Session import Session
# from .WRAPPERS import LogsManager, Logger, VarsManager, Flow, DataConnection



# import math
from qtpy.QtCore import QMarginsF
from qtpy.QtCore import QRectF, QPointF, Qt
from qtpy.QtGui import QPainter, QColor, QRadialGradient, QPainterPath, QPen
from qtpy.QtWidgets import QGraphicsPathItem, QGraphicsItem, QStyleOptionGraphicsItem

from ...GUIBase import GUIBase
from ...utils import sqrt
from ...utils import pythagoras


class ConnectionItem(GUIBase, QGraphicsPathItem):
    """The GUI representative for a connection. The classes ExecConnectionItem and DataConnectionItem will be ready
    for reimplementation later, so users can add GUI for the enhancements of DataConnection and ExecConnection,
    like input fields for weights."""

    def __init__(self, connection, session_design):
        GUIBase.__init__(self, representing_component=connection)
        QGraphicsPathItem.__init__(self)

        self.setAcceptHoverEvents(True)

        self.connection = connection

        out = self.connection.out
        inp = self.connection.inp
        out_node = out.node
        inp_node = inp.node
        self.out_item = out_node.port_item(out)
        self.inp_item = inp_node.port_item(inp)

        self.session_design = session_design
        self.session_design.flow_theme_changed.connect(self.recompute)
        self.session_design.performance_mode_changed.connect(self.recompute)

        # for rendering flow pictures
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

        self.recompute()

    def recompute(self):
        """Updates scene position and recomputes path, pen and gradient"""

        # position
        self.setPos(self.out_pos())

        # path
        self.setPath(
            self.connection_path(
                QPointF(0, 0),
                self.inp_pos()-self.scenePos()
            )
        )

        # pen
        pen = self.get_pen()

        # brush
        self.setBrush(Qt.NoBrush)

        #   gradient
        if self.session_design.performance_mode == 'pretty':
            c = pen.color()
            w = self.path().boundingRect().width()
            h = self.path().boundingRect().height()
            gradient = QRadialGradient(
                self.boundingRect().center(),
                pythagoras(w, h) / 2
            )

            c_r = c.red()
            c_g = c.green()
            c_b = c.blue()

            # this offset will be 1 if inp.x >> out.x and 0 if inp.x < out.x
            # hence, no fade for the gradient if the connection goes backwards
            offset_mult: float = max(
                0,
                min(
                    (self.inp_pos().x() - self.out_pos().x()) / 200,
                    1
                )
            )

            # and if the input is very far away from the output, decrease the gradient fade so the connection
            # doesn't fully disappear at the ends and stays visible
            if self.inp_pos().x() > self.out_pos().x():
                offset_mult = min(
                    offset_mult,
                    2000 / (self.dist(self.inp_pos(), self.out_pos()))
                )
                # zucker.

            gradient.setColorAt(0.0, QColor(c_r, c_g, c_b, 255))
            gradient.setColorAt(0.75, QColor(c_r, c_g, c_b, 255 - round(55 * offset_mult)))
            gradient.setColorAt(0.95, QColor(c_r, c_g, c_b, 255 - round(255 * offset_mult)))

            pen.setBrush(gradient)

        self.setPen(pen)
    
    def out_pos(self) -> QPointF:
        """The current global scene position of the pin of the output port"""

        return self.out_item.pin.get_scene_center_pos()
    
    def inp_pos(self) -> QPointF:
        """The current global scene position of the pin of the input port"""

        return self.inp_item.pin.get_scene_center_pos()

    def set_highlighted(self, b: bool):
        pen: QPen = self.pen()

        if b:
            pen.setWidthF(self.pen_width() * 2)
        else:
            pen.setWidthF(self.pen_width())
            self.recompute()

        self.setPen(pen)

    def get_pen(self) -> QPen:
        pass

    def pen_width(self) -> int:
        pass

    def flow_theme(self):
        return self.session_design.flow_theme

    def hoverEnterEvent(self, event):
        self.set_highlighted(True)
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self.set_highlighted(False)
        super().hoverLeaveEvent(event)

    @staticmethod
    def dist(p1: QPointF, p2: QPointF) -> float:
        """Returns the diagonal distance between the points using pythagoras"""

        dx = p2.x()-p1.x()
        dy = p2.y()-p1.y()
        return sqrt((dx**2) + (dy**2))


    @staticmethod
    def connection_path(p1: QPointF, p2: QPointF) -> QPainterPath:
        """Returns the painter path for drawing the connection, using the usual cubic connection path by default"""

        return default_cubic_connection_path(p1, p2)


class ExecConnectionItem(ConnectionItem):

    def pen_width(self):
        return self.flow_theme().exec_conn_width

    def get_pen(self):
        theme = self.flow_theme()
        pen = QPen(theme.exec_conn_color, theme.exec_conn_width)
        pen.setStyle(theme.exec_conn_pen_style)
        pen.setCapStyle(Qt.RoundCap)
        return pen


class DataConnectionItem(ConnectionItem):

    def pen_width(self):
        return self.flow_theme().data_conn_width

    def get_pen(self):
        theme = self.flow_theme()
        pen = QPen(theme.data_conn_color, theme.data_conn_width)
        pen.setStyle(theme.data_conn_pen_style)
        pen.setCapStyle(Qt.RoundCap)
        return pen


def default_cubic_connection_path(p1: QPointF, p2: QPointF):
    """Returns the nice looking QPainterPath from p1 to p2"""

    path = QPainterPath()

    path.moveTo(p1)

    dx = p2.x() - p1.x()
    adx = abs(dx)
    dy = p2.y() - p1.y()
    ady = abs(dy)
    distance = sqrt((dx ** 2) + (dy ** 2))
    x1, y1 = p1.x(), p1.y()
    x2, y2 = p2.x(), p2.y()

    if ((x1 < x2 - 30) or distance < 100) and (x1 < x2):
        # STANDARD FORWARD
        path.cubicTo(x1 + ((x2 - x1) / 2), y1,
                     x1 + ((x2 - x1) / 2), y2,
                     x2, y2)
    elif x2 < x1 - 100 and adx > ady * 2:
        # STRONG BACKWARDS
        path.cubicTo(x1 + 100 + (x1 - x2) / 10, y1,
                     x1 + 100 + (x1 - x2) / 10, y1 + (dy / 2),
                     x1 + (dx / 2), y1 + (dy / 2))
        path.cubicTo(x2 - 100 - (x1 - x2) / 10, y2 - (dy / 2),
                     x2 - 100 - (x1 - x2) / 10, y2,
                     x2, y2)
    else:
        # STANDARD BACKWARDS
        path.cubicTo(x1 + 100 + (x1 - x2) / 3, y1,
                     x2 - 100 - (x1 - x2) / 3, y2,
                     x2, y2)
    
    return path



import bisect
import enum
import json
import os
from math import sqrt
from waiting import wait
from typing import List, Dict

from ryvencore.utils import serialize, deserialize
from qtpy.QtCore import QPointF


class Container:
    """used for threading; accessed from multiple threads"""

    def __init__(self):
        self.payload = None
        self.has_been_set = False

    def set(self, val):
        self.payload = val
        self.has_been_set = True

    def is_set(self):
        return self.has_been_set


def wait_until(func):
    return wait(func, sleep_seconds=0.001)


def pythagoras(a, b):
    return sqrt(a ** 2 + b ** 2)


def get_longest_line(s: str):
    lines = s.split('\n')
    lines = [line.replace('\n', '') for line in lines]
    longest_line_found = ''
    for line in lines:
        if len(line) > len(longest_line_found):
            longest_line_found = line
    return line


def shorten(s: str, max_chars: int, line_break: bool = False):
    """Ensures, that a given string does not exceed a given max length. If it would, its cut in the middle."""
    l = len(s)
    if l > max_chars:
        insert = ' . . . '
        if line_break:
            insert = '\n'+insert+'\n'
        insert_length = len(insert)
        left = s[:round((max_chars-insert_length)/2)]
        right = s[round(l-((max_chars-insert_length)/2)):]
        return left+insert+right
    else:
        return s


def pointF_mapped(p1, p2):
    """adds the floating part of p2 to p1"""
    p2.setX(p1.x() + p2.x()%1)
    p2.setY(p1.y() + p2.y()%1)
    return p2

def points_dist(p1, p2):
    return sqrt(abs(p1.x() - p2.x())**2 + abs(p1.y() - p2.y())**2)

def middle_point(p1, p2):
    return QPointF((p1.x() + p2.x())/2, (p1.y() + p2.y())/2)


class MovementEnum(enum.Enum):
    # this should maybe get removed later
    mouse_clicked = 1
    position_changed = 2
    mouse_released = 3


def change_svg_color(filepath: str, color_hex: str):
    # doesnt seem to work properly yet :(

    from qtpy.QtSvg import QSvgRenderer
    from qtpy.QtGui import QPixmap, QPainter
    from qtpy.QtCore import Qt

    with open(filepath) as f:
        data = f.read()
    data = data.replace('fill:#000000', 'fill:'+color_hex)
    with open(filepath, 'w') as f:
        f.write(data)

    svg_renderer = QSvgRenderer(filepath)
    pix = QPixmap(svg_renderer.defaultSize())
    pix.fill(Qt.transparent)
    pix_painter = QPainter(pix)
    svg_renderer.render(pix_painter)

    return pix



    # """
    # Changes the color of an SVG image and returns a QPixmap
    #
    # https://stackoverflow.com/questions/15123544/change-the-color-of-an-svg-in-qt
    # """
    #
    # from qtpy.QtGui import Qt, QPainter
    # from qtpy.QtXml import QDomDocument, QDomElement
    # from qtpy.QtSvg import QSvgRenderer
    # from qtpy.QtGui import QPixmap
    #
    #
    # def change_svg_color__set_attr_recur(elem: QDomElement, strtagname: str, strattr: str, strattrval: str):
    #
    #     # if it has the tag name then overwrite desired attribute
    #     if elem.tagName() == strtagname:
    #         elem.setAttribute(strattr, strattrval)
    #
    #     # loop all children
    #     for i in range(elem.childNodes().count()):
    #         if not elem.childNodes().at(i).isElement():
    #             continue
    #
    #         change_svg_color__set_attr_recur(elem.childNodes().at(i).toElement(), strtagname, strattr, strattrval)
    #
    #
    # # open svg resource load contents to qbytearray
    # f = open(filepath)
    # data = f.read()
    # f.close()
    # # load svg contents to xml document and edit contents
    # doc = QDomDocument()
    # doc.setContent(data)
    # # recursively change color
    # change_svg_color__set_attr_recur(doc.documentElement(), 'path', 'fill', color_hex)
    # # create svg renderer with edited contents
    # svg_renderer = QSvgRenderer(doc.toByteArray())
    # # create pixmap target (could be a QImage)
    # pix = QPixmap(svg_renderer.defaultSize())
    # pix.fill(Qt.transparent)
    # # create painter to act over pixmap
    # pix_painter = QPainter(pix)
    # # use renderer to render over painter which paints on pixmap
    # svg_renderer.render(pix_painter)
    #
    # return pix


def translate_project(project: Dict) -> Dict:
    """
    Transforms a v3.0 project file into something that can be loaded in v3.1,
    i.e. turns macros into scripts and removes macro nodes from the flows.
    """

    new_project = project.copy()

    # turn macros into scripts

    fixed_scripts = []

    for script in (project['macro scripts']+project['scripts']):

        new_script = script.copy()

        # remove macro nodes
        new_nodes, removed_node_indices = remove_macro_nodes(script['flow']['nodes'])
        new_script['flow']['nodes'] = new_nodes

        # fix connections
        new_script['flow']['connections'] = fix_connections(script['flow']['connections'], removed_node_indices)

        fixed_scripts.append(new_script)

    del new_project['macro scripts']
    new_project['scripts'] = fixed_scripts

    return new_project


def remove_macro_nodes(nodes):
    """
    removes all macro nodes from the nodes list and returns the new list as well as the indices of the removed nodes
    """

    new_nodes = []
    removed_node_indices = []

    for n_i in range(len(nodes)):
        node = nodes[n_i]

        if node['identifier'] in ('BUILTIN_MacroInputNode', 'BUILTIN_MacroOutputNode') or \
                node['identifier'].startswith('MACRO_NODE_'):
            removed_node_indices.append(n_i)
        else:
            new_nodes.append(node)

    return new_nodes, removed_node_indices


def fix_connections(connections: Dict, removed_node_indices: List) -> List:
    """
    removes connections to removed nodes and fixes node indices of the other ones
    """

    import bisect

    new_connections = []

    for conn in connections:
        if conn['parent node index'] in removed_node_indices or conn['connected node'] in removed_node_indices:
            # remove connection
            continue
        else:
            # fix node indices
            pni = conn['parent node index']
            cni = conn['connected node']

            #   calculate the number of removed nodes with indices < pni | cni
            num_smaller_removed_pni = bisect.bisect_left(removed_node_indices, pni)
            num_smaller_removed_cni = bisect.bisect_left(removed_node_indices, cni)

            c = conn.copy()

            #   decrease indices accordingly
            c['parent node index'] = pni - num_smaller_removed_pni
            c['connected node'] = cni - num_smaller_removed_cni

            new_connections.append(c)

    return new_connections



# list widgets
from .VariablesListWidget import VariablesListWidget as VarsList
from .ScriptsListWidget import ScriptsListWidget as ScriptsList
from ..flows.node_list_widget.NodeListWidget import NodeListWidget

# logging
from .LogWidget import LogWidget



"""
This file contains the implementations of undoable actions for FlowView.

Why not in Flow?
-> Because there are actions that only regard the GUI and are therefore unnoticed by the Flow but still undoable, like
adding a DrawingObject.

The main characteristic of these implementations is that the communication to the Flow is down through the threading
interface, because these actions and the Flow might live in different threads.
"""


from qtpy.QtCore import QObject, QPointF
from qtpy.QtWidgets import QUndoCommand

from .drawings.DrawingObject import DrawingObject
from .nodes.NodeItem import NodeItem


class FlowUndoCommand(QObject, QUndoCommand):
    """
    The main difference to normal QUndoCommands is the activate feature. This allows the flow widget to add the
    undo command to the undo stack before redo() is called. This is important since some of these commands can cause
    other commands to be added while they are performing redo(), so to prevent those commands to be added to the
    undo stack before the parent command, it is here blocked at first.
    """

    def __init__(self, flow_view):

        self.flow_view = flow_view
        self.flow = flow_view.flow
        self._activated = False

        QObject.__init__(self)
        QUndoCommand.__init__(self)

    def activate(self):
        self._activated = True
        self.redo()

    def redo(self) -> None:
        if not self._activated:
            return
        else:
            self.redo_()

    def undo(self) -> None:
        self.undo_()

    def redo_(self):
        """subclassed"""
        pass

    def undo_(self):
        """subclassed"""
        pass


    def call(self, target_method, args: tuple, then=None):
        """Calls the backend thread to execute some method and runs 'then' with the response if provided"""

        # disable threading for now
        ret = target_method(*args)
        if then is not None:
            then(ret)

        # ret = self.flow_view.thread_interface.run(target_method, args)
        #
        # if then is not None:
        #     then(ret)


class MoveComponents_Command(FlowUndoCommand):
    def __init__(self, flow_view, items_list, p_from, p_to):
        super(MoveComponents_Command, self).__init__(flow_view)

        self.items_list = items_list
        self.p_from = p_from
        self.p_to = p_to
        self.last_item_group_pos = p_to

    def undo_(self):
        items_group = self.items_group()
        items_group.setPos(self.p_from)
        self.last_item_group_pos = items_group.pos()
        self.destroy_items_group(items_group)

    def redo_(self):
        items_group = self.items_group()
        items_group.setPos(self.p_to - self.last_item_group_pos)
        self.destroy_items_group(items_group)


    def items_group(self):
        return self.flow_view.scene().createItemGroup(self.items_list)

    def destroy_items_group(self, items_group):
        self.flow_view.scene().destroyItemGroup(items_group)


class PlaceNode_Command(FlowUndoCommand):

    def __init__(self, flow_view, node_class, pos):
        super().__init__(flow_view)

        self.node_class = node_class
        self.node = None
        self.item_pos = pos

        self.flow_view.node_placed.connect(self.node_placed_in_flow_view)

    def undo_(self):
        self.call(self.flow.remove_node, (self.node,))

    def redo_(self):
        if self.node:
            self.call(self.flow.add_node, (self.node,))
        else:
            self.flow_view.node_placed.connect(self.node_placed_in_flow_view)
            self.call(self.flow.create_node, (self.node_class,))

        # --> node_placed_in_flow_view()

    def node_placed_in_flow_view(self, node):
        self.flow_view.node_placed.disconnect(self.node_placed_in_flow_view)
        self.node = node


class PlaceDrawing_Command(FlowUndoCommand):
    def __init__(self, flow_view, posF, drawing):
        super().__init__(flow_view)

        self.drawing = drawing
        self.drawing_obj_place_pos = posF
        self.drawing_obj_pos = self.drawing_obj_place_pos

    def undo_(self):
        # The drawing_obj_pos is not anymore the drawing_obj_place_pos because after the
        # drawing object was completed, its actual position got recalculated according to all points and differs from
        # the initial pen press pos (=drawing_obj_place_pos). See DrawingObject.finished().

        self.drawing_obj_pos = self.drawing.pos()

        self.flow_view.remove_component(self.drawing)

    def redo_(self):
        self.flow_view.add_drawing(self.drawing, self.drawing_obj_pos)


class RemoveComponents_Command(FlowUndoCommand):

    def __init__(self, flow_view, items):
        super().__init__(flow_view)

        self.items = items
        self.broken_connections = []  # the connections that go beyond the removed nodes and need to be restored in undo
        self.internal_connections = set()

        self.node_items = []
        self.nodes = []
        self.drawings = []
        for i in self.items:
            if isinstance(i, NodeItem):
                self.node_items.append(i)
                self.nodes.append(i.node)
            elif isinstance(i, DrawingObject):
                self.drawings.append(i)

        for n in self.nodes:
            for i in n.inputs:
                for c in i.connections:
                    cp = c.out
                    cn = cp.node
                    if cn not in self.nodes:
                        self.broken_connections.append(c)
                    else:
                        self.internal_connections.add(c)
            for o in n.outputs:
                for c in o.connections:
                    cp = c.inp
                    cn = cp.node
                    if cn not in self.nodes:
                        self.broken_connections.append(c)
                    else:
                        self.internal_connections.add(c)

    def undo_(self):

        # add nodes
        for n in self.nodes:
            self.call(self.flow.add_node, (n,))

        # add drawings
        for d in self.drawings:
            self.flow_view.add_drawing(d)

        # add connections
        self.restore_broken_connections()
        self.restore_internal_connections()

    def redo_(self):

        # remove connections
        self.remove_broken_connections()
        self.remove_internal_connections()

        # remove nodes
        for n in self.nodes:
            self.call(self.flow.remove_node, (n,))

        # remove drawings
        for d in self.drawings:
            self.flow_view.remove_drawing(d)

    def restore_internal_connections(self):
        for c in self.internal_connections:
            self.call(self.flow.add_connection, (c,))

    def remove_internal_connections(self):
        for c in self.internal_connections:
            self.call(self.flow.remove_connection, (c,))

    def restore_broken_connections(self):
        for c in self.broken_connections:
            self.call(self.flow.add_connection, (c,))

    def remove_broken_connections(self):
        for c in self.broken_connections:
            self.call(self.flow.remove_connection, (c,))


class ConnectPorts_Command(FlowUndoCommand):

    def __init__(self, flow_view, out, inp):
        super().__init__(flow_view)

        # CAN ALSO LEAD TO DISCONNECT INSTEAD OF CONNECT!!

        self.out = out
        self.inp = inp
        self.connection = None
        self.connecting = True

        for c in self.out.connections:
            if c.inp == self.inp:
                self.connection = c
                self.connecting = False


    def undo_(self):
        if self.connecting:
            # remove connection
            self.call(self.flow.remove_connection, (self.connection,))
        else:
            # recreate former connection
            self.call(self.flow.add_connection, (self.connection,))

    def redo_(self):
        if self.connecting:
            if self.connection:
                self.call(self.flow.add_connection, (self.connection,))
            else:
                # connection hasn't been created yet
                self.flow.connection_added.connect(self.connection_created)
                self.call(self.flow.connect_nodes, (self.out, self.inp))
        else:
            # remove existing connection
            self.call(self.flow.remove_connection, (self.connection,))

    def connection_created(self, c):
        self.flow.connection_added.disconnect(self.connection_created)
        self.connection = c




class Paste_Command(FlowUndoCommand):

    def __init__(self, flow_view, data, offset_for_middle_pos):
        super().__init__(flow_view)

        self.data = data
        self.modify_data_positions(offset_for_middle_pos)
        self.pasted_components = None


    def modify_data_positions(self, offset):
        """adds the offset to the components' positions in data"""

        for node in self.data['nodes']:
            node['pos x'] = node['pos x'] + offset.x()
            node['pos y'] = node['pos y'] + offset.y()
        for drawing in self.data['drawings']:
            drawing['pos x'] = drawing['pos x'] + offset.x()
            drawing['pos y'] = drawing['pos y'] + offset.y()

    def redo_(self):
        if self.pasted_components is None:
            self.pasted_components = {}

            # create components
            self.create_drawings()
            self.call(self.flow.create_nodes_from_data, (self.data['nodes'],),
                      then=self.nodes_created)
            # --> nodes_created()
        else:
            self.add_existing_components()

    def undo_(self):
        # remove components and their items from flow
        for n in self.pasted_components['nodes']:
            self.call(self.flow.remove_node, (n,))
        for c in self.pasted_components['connections']:
            self.call(self.flow.remove_connection, (c,))
        for d in self.pasted_components['drawings']:
            self.flow_view.remove_drawing(d)

    def add_existing_components(self):
        # add existing components and items to flow
        for n in self.pasted_components['nodes']:
            self.call(self.flow.add_node, (n,))
        for c in self.pasted_components['connections']:
            self.call(self.flow.add_connection, (c,))
        for d in self.pasted_components['drawings']:
            self.flow_view.add_drawing(d)

        self.select_new_components_in_view()

    def select_new_components_in_view(self):
        self.flow_view.clear_selection()
        for d in self.pasted_components['drawings']:
            d: DrawingObject
            d.setSelected(True)
        for n in self.pasted_components['nodes']:
            n: NodeItem
            ni: NodeItem = self.flow_view.node_items[n]
            ni.setSelected(True)




    def nodes_created(self, nodes):
        self.pasted_components['nodes'] = nodes
        self.call(self.flow.connect_nodes_from_data, (nodes, self.data['connections']),
                  then=self.connections_created)
        # --> connections_created()

    def connections_created(self, connections):
        self.pasted_components['connections'] = connections

        self.select_new_components_in_view()

    def create_drawings(self):
        drawings = []
        for d in self.data['drawings']:
            new_drawing = self.flow_view.create_drawing(d)
            self.flow_view.add_drawing(new_drawing, posF=QPointF(d['pos x'], d['pos y']))
            drawings.append(new_drawing)
        self.pasted_components['drawings'] = drawings



from qtpy.QtWidgets import QWidget, QApplication
from qtpy.QtCore import QObject, Signal, Qt

from ..GUIBase import GUIBase
from ..SessionThreadInterface import SessionThreadInterface_Backend
from ..Design import Design
from ..flows.FlowView import FlowView
from ..utils import translate_project

from ryvencore import Session as RC_Session, Script
from ryvencore.Base import Base

from typing import Dict, List


class Session(RC_Session, QObject):

    flow_view_created = Signal(object, object)

    def __init__(
            self,
            gui_parent: QWidget = None,
    ):
        QObject.__init__(self)

        RC_Session.__init__(self, gui=True)

        # flow views
        self.flow_views = {}  # {Script : FlowView}

        # gui hierarchy
        self.gui_parent = gui_parent

        # notice: true treading is currently not supported
        self.threading_bridge__backend = SessionThreadInterface_Backend()
        self.threading_bridge__frontend = self.threading_bridge__backend.frontend
        # self.threading_bridge__frontend.moveToThread(gui_parent.thread() if threaded else self.thread())

        # set complete_data function (needs threading_bridge)
        Base.complete_data_function = GUIBase.get_complete_data_function(self)

        # design
        app = QApplication.instance()
        app.setAttribute(Qt.AA_UseHighDpiPixmaps)
        Design.register_fonts()
        self.design = Design()


    def load(self, project: Dict) -> List[Script]:

        if 'macro scripts' in project:
            # backwards compatibility
            project = translate_project(project)

        return super().load(project)


    def create_script(self, title: str = None, create_default_logs=True,
                      data: dict = None,
                      flow_view_size: list = None) -> Script:
        """Creates and returns a new script, and builds the flow view.
        If a data is provided the title parameter will be ignored."""

        script = RC_Session.create_script(self, title=title, create_default_logs=create_default_logs, data=data)

        # self.new_script_created.emit(script)

        self._build_flow_view(script, flow_view_size)

        return script


    def _build_flow_view(self, script, view_size):
        """Builds the flow view in the GUI thread"""

        script_data = script.init_data

        flow_view_data = None
        if script_data is not None:
            if 'flow view' not in script_data:  # backwards compatibility
                flow_view_data = script_data['flow']
            else:
                flow_view_data = script_data['flow view']

        flow_view = self.threading_bridge__backend.run(
            FlowView, (
                self,
                script,
                script.flow,
                flow_view_data,
                view_size,
                self.gui_parent,
            )
        )

        self.flow_views[script] = flow_view
        self.flow_view_created.emit(script, flow_view)

        return flow_view

    def _flow_view_created(self, view):
        self.temp_data = view

    def _script_from_title(self, title: str) -> Script:
        for s in self.scripts:
            if s.title == title:
                return s
        return None




# set package path (for resources etc.)
import os
from .src.GlobalAttributes import Location
Location.PACKAGE_PATH = os.path.normpath(os.path.dirname(__file__))

# set ryvencore gui mode
os.environ['RC_MODE'] = 'gui'

# import backend wrapper
from .src.core_wrapper import *

# import frontend
from .src.flows.nodes.WidgetBaseClasses import MWB, IWB
from .src.flows.connections.ConnectionItem import DataConnectionItem, ExecConnectionItem
from .src.conv_gui import *

# expose ryvencore
import ryvencore
